from copy import deepcopy

from payton.math.functions import distance, mid_point
from payton.scene.geometry import Mesh


def merge_mesh(mesh1: Mesh, mesh2: Mesh) -> Mesh:
    """Merge two meshes into a single mesh

    Keyword arguments:
    mesh1 -- First mesh
    mesh2 -- Second mesh
    """
    m1_vertices = list(mesh1.absolute_vertices())
    m2_vertices = list(mesh2.absolute_vertices())
    mesh = Mesh()
    mesh.material = deepcopy(mesh1.material)
    mesh._vertices = m1_vertices + m2_vertices
    mesh._vertex_colors = mesh1._vertex_colors + mesh2._vertex_colors
    mesh._vertex_count = len(mesh._vertices)
    mesh._texcoords = mesh1._texcoords + mesh2._texcoords
    mesh._normals = mesh1._normals + mesh2._normals
    len_v = len(mesh1._vertices)
    m2_indices = [[i[0] + len_v, i[1] + len_v, i[2] + len_v] for i in mesh2._indices]
    mesh._indices = mesh1._indices + m2_indices
    mesh.material._indices = mesh._indices
    mesh._calc_bounds()
    mesh.refresh()
    return mesh


def subdivide(original: Mesh) -> Mesh:
    """Subdivide the mesh polygons by given steps

    Keyword arguments:
    mesh -- Original mesh
    """
    new = Mesh()
    has_texcoords = len(original._texcoords) == len(original._vertices)
    new.clear_triangles()

    for indice in original._indices:
        v1 = original._vertices[indice[0]]
        v2 = original._vertices[indice[1]]
        v3 = original._vertices[indice[2]]

        a = v1
        a_i = indice[0]
        b = v2
        b_i = indice[1]
        c = v3
        c_i = indice[2]

        if distance(v2, v3) > distance(v1, v2):
            a = v2
            a_i = indice[1]
            b = v3
            b_i = indice[2]
            c = v1
            c_i = indice[0]

        if distance(v3, v1) > distance(v2, v3):
            a = v3
            a_i = indice[2]
            b = v1
            b_i = indice[0]
            c = v2
            c_i = indice[1]

        vc = mid_point(a, b)
        n1 = original._normals[indice[0]]
        texcoords_1 = None
        texcoords_2 = None
        if has_texcoords:
            t1 = original._texcoords[a_i]
            t2 = original._texcoords[b_i]
            t3 = original._texcoords[c_i]
            t1.append(0)
            t2.append(0)
            t3.append(0)

            tc = mid_point(t1, t2)

            texcoords_1 = [t1[:2], tc[:2], t3[:2]]
            texcoords_2 = [t3[:2], tc[:2], t2[:2]]

        new.add_triangle(
            vertices=[a, vc, c],
            normals=[n1, n1, n1],
            texcoords=texcoords_1,
        )
        new.add_triangle(
            vertices=[c, vc, b],
            normals=[n1, n1, n1],
            texcoords=texcoords_2,
        )
    return new



import math
from copy import deepcopy
from functools import lru_cache
from typing import Iterable, List, Tuple

import numpy as np
import pyrr

from payton.math.matrix import IDENTITY_MATRIX, Matrix
from payton.math.vector import Vector3D


def to_4(v: Iterable, fill: float = 1.0) -> Vector3D:
    res = list(v)
    for _ in range(4 - len(res)):
        res.append(fill)
    return res


def sub_vector(v1: Vector3D, v2: Vector3D) -> Vector3D:
    """Substract vector V1 from vector V2 and return resulting Vector.

    Keyword arguments:
    v1 -- Vector 1
    v2 -- Vector 2
    """
    return [v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2]]


def cross_product(v1: Vector3D, v2: Vector3D) -> Vector3D:
    """Return the cross product vector of V1 and V2.

    Keyword arguments:
    v1 -- Vector 1
    v2 -- Vector 2
    """
    return [
        v1[1] * v2[2] - v1[2] * v2[1],
        v1[2] * v2[0] - v1[0] * v2[2],
        v1[0] * v2[1] - v1[1] * v2[0],
    ]


def vector_norm(v: Vector3D) -> float:
    """Calculate and return the Norm of a given Vector."""
    return math.sqrt((v[0] * v[0]) + (v[1] * v[1]) + (v[2] * v[2]))


def normalize_vector(v: Vector3D) -> Vector3D:
    """
    Normalize given vector and output normalized vector.

    Keyword Arguments:
    v -- Vector to normalize
    """
    vn = vector_norm(v)
    if vn == 0.0:
        return v
    invlen = 1.0 / vn
    return [v[0] * invlen, v[1] * invlen, v[2] * invlen]


def plane_normal(v1: Vector3D, v2: Vector3D, v3: Vector3D) -> Vector3D:
    """
    Calculate the normal of the plane defined by v1, v2, v3 Vectors in space.

    Keyword arguments:
    v1 -- V1 Vector of the plane (Corner of the triangle in the given plane)
    v2 -- V2 Vector of the plane (Corner of the triangle in the given plane)
    v3 -- V3 Vector of the plane (Corner of the triangle in the given plane)
    """
    return normalize_vector(cross_product(sub_vector(v2, v1), sub_vector(v3, v1)))


def vector_transform(v: Vector3D, matrix: Matrix) -> Vector3D:
    """
    Transform the given Vector into given Matrix coordinates.

    Keyword arguments:
    v -- Vector to transform
    matrix -- Matrix for the transformation
    """
    mx = matrix[0]
    my = matrix[1]
    mz = matrix[2]
    mw = matrix[3]

    rx = v[0] * mx[0] + v[1] * my[0] + v[2] * mz[0]
    ry = v[0] * mx[1] + v[1] * my[1] + v[2] * mz[1]
    rz = v[0] * mx[2] + v[1] * my[2] + v[2] * mz[2]

    return [rx + mw[0], ry + mw[1], rz + mw[2]]


def invert_vector(v: Vector3D) -> Vector3D:
    """
    Invert - Negate the given vector.

    Keyword arguments:
    v -- Vector to invert
    """
    return [-v[0], -v[1], -v[2]]


def min_max(vlist: List[Vector3D]) -> Tuple[Vector3D, Vector3D]:
    """
    Calculate the smallest and biggest components of the given vector list.

    Returns the result as Tuple of Vectors as minVector - maxVector

    Example:
        vlist = [
            Vector(-2, 3, 4),
            Vector(-3, -5, 9)
        ]

        min_max(vlist)

    would return:

        ([-3, -5, 4], [-2, 3, 9])


    Keyword arguments:
    vlist -- List of vectors
    """
    x_list = [item[0] for item in vlist]
    y_list = [item[1] for item in vlist]
    z_list = [item[2] for item in vlist]
    return (
        [min(x_list), min(y_list), min(z_list)],
        [max(x_list), max(y_list), max(z_list)],
    )


def scale_vector(v: Vector3D, factor: float) -> Vector3D:
    """
    Scale (multiply each component) of the given Vector and return as a result.

    Keyword arguments:
    v -- Vector to scale
    factor -- factor to multiply
    """
    return [v[0] * factor, v[1] * factor, v[2] * factor]


def add_vectors(v1: Vector3D, v2: Vector3D) -> Vector3D:
    """
    Add two vectors.

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    return [v1[0] + v2[0], v1[1] + v2[1], v1[2] + v2[2]]


def dot_product(v1: Vector3D, v2: Vector3D) -> float:
    """
    Return vector dot product of v1 and v2.

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


def vector_angle(v1: Vector3D, v2: Vector3D) -> float:
    """
    Calculate the angle between two given vectors.

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    v1_n = normalize_vector(v1)
    v2_n = normalize_vector(v2)
    return math.acos(dot_product(v1_n, v2_n) / (vector_norm(v1_n) * vector_norm(v2_n)))


def mid_point(v1: Vector3D, v2: Vector3D) -> Vector3D:
    """
    Calculate the middle point between v1 and v2 vectors.

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    diff = sub_vector(v2, v1)
    leng = vector_norm(diff)
    diff = normalize_vector(diff)
    diff = scale_vector(diff, leng / 2.0)
    return add_vectors(diff, v1)


def distance(v1: Vector3D, v2: Vector3D) -> float:
    """
    Calculate the distance between vectors v1 and v2.

    Keyword arguments:
    v1 -- First vector
    v2 -- Second vector
    """
    diff = sub_vector(v2, v1)
    return vector_norm(diff)


def rotate_around_z(v: Vector3D, angle: float) -> Vector3D:
    """Rotate the given vector around Z axis by given degrees.

    Keyword arguments:
    v -- Vector to rotate
    angle -- Angle of rotation in Radians
    """
    s = math.sin(angle)
    c = math.cos(angle)
    return [v[0] * c + v[1] * s, v[1] * c - v[0] * s, v[2]]


def create_rotation_matrix_raw(axis: Vector3D, angle: float) -> Matrix:
    """
    Create a rotation matrix as Numpy array.

    If you transform a Vector by the rotation Matrix, you will be rotating
    the given Vector according to the rotation defined by the matrix.

    If you multiply a matrix by the rotation matrix, resulting matrix will be rotated
    by the given rotation matrix.

    Keyword arguments:
    axis -- Rotation axis
    angle -- Angle of rotation in Radians
    """
    sin = math.sin(angle)
    cos = math.cos(angle)
    m_cos = 1 - cos
    axis = normalize_vector(axis)
    return [
        [
            (m_cos * axis[0] * axis[0]) + cos,
            (m_cos * axis[0] * axis[1]) - (axis[2] * sin),
            (m_cos * axis[2] * axis[0]) + (axis[1] * sin),
            0,
        ],
        [
            (m_cos * axis[0] * axis[1]) + (axis[2] * sin),
            (m_cos * axis[1] * axis[1]) + cos,
            (m_cos * axis[1] * axis[2]) - (axis[0] * sin),
            0,
        ],
        [
            (m_cos * axis[2] * axis[0]) - (axis[1] * sin),
            (m_cos * axis[1] * axis[2]) + (axis[0] * sin),
            (m_cos * axis[2] * axis[2]) + cos,
            0,
        ],
        [0.0, 0.0, 0.0, 1.0],
    ]


def create_rotation_matrix(axis: Vector3D, angle: float) -> np.ndarray:
    """
    Create rotation matrix as a numpy array.

    Numpy array is a well packed proper memory array that fits well into
    graphics card. Therefore, we are using Numpy Arrays to push stuff to
    the graphics card.

    If you transform a Vector by the rotation Matrix, you will be rotating
    the given Vector according to the rotation defined by the matrix.

    If you multiply a matrix by the rotation matrix, resulting matrix will be rotated
    by the given rotation matrix.

    Keyword arguments:
    axis -- Rotation axis
    angle -- Angle of rotation in Radians.
    """
    result = create_rotation_matrix_raw(axis, angle)
    return np.array(result, dtype=np.float32)


def scale_matrix(x: float, y: float, z: float) -> np.ndarray:
    """
    Create a scale matrix.

    If you multiply an objects matrix with a scale matrix, it will scale / resize
    the coordinates relative to the object. As a result, the object will look resized.

    Keyword arguments:
    x -- X Scale Factor
    y -- Y Scale Factor
    z -- Z Scale Factor
    """
    result = deepcopy(IDENTITY_MATRIX)
    result = [[x, 0, 0, 0], [0, y, 0, 0], [0, 0, z, 0], [0, 0, 0, 1]]
    return np.array(result, dtype=np.float32)


@lru_cache(maxsize=512)
def ortho(left: float, right: float, bottom: float, top: float) -> np.ndarray:
    """
    Create an orthographic projection matrix as a Numpy Array.

    Keyword arguments:
    left -- Viewport left side (-x) coordinate
    right -- Viewport right side (+x) coordinate
    bottom -- Viewport bottom side (+y) coordinate
    top -- Viewport top side (-y) coordinate
    """
    result = [
        [2 / (right - left), 0, 0, 0],
        [0, 2 / (top - bottom), 0, 0],
        [0, 0, -1, 0],
        [-(right + left) / (right - left), -(top + bottom) / (top - bottom), 0, 1],
    ]
    return np.array(result, dtype=np.float32)


def cubemap_projection_matrices(from_point: Vector3D, far_plane: float) -> List[np.ndarray]:
    """
    Create the required Cubemap projection matrices.

    This method is suitable for generating a Shadow Map.

    Simply speaking, this method generates 6 different camera matrices from the center of
    an imaginary cube and covers all surfaces without conflicting.

    Keyword arguments;
    from_point -- Imaginary camera location
    far_plane -- How far the camera is capable of seeing. (Effects performance!)
    """

    def a2np(a: List[float]) -> np.ndarray:
        return np.array(a, dtype=np.float32)

    shadow_proj = pyrr.matrix44.create_perspective_projection(90.0, 1.0, 0.01, far_plane, np.float32)
    lightpos = np.array(list(from_point), dtype=np.float32)[:3]

    nx = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([-1.0, 0, 0]),
            dtype=np.float32,
        ),
        a2np([0, -1.0, 0]),
        dtype=np.float32,
    )
    px = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([1, 0, 0]),
            dtype=np.float32,
        ),
        a2np([0, -1.0, 0]),
        dtype=np.float32,
    )
    ny = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([0, -1, 0]),
            dtype=np.float32,
        ),
        a2np([0, 0, -1.0]),
        dtype=np.float32,
    )
    py = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([0, 1, 0]),
            dtype=np.float32,
        ),
        a2np([0, 0, 1.0]),
        dtype=np.float32,
    )
    pz = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([0, 0, 1]),
            dtype=np.float32,
        ),
        a2np([0, -1.0, 0]),
        dtype=np.float32,
    )
    nz = pyrr.matrix44.create_look_at(
        lightpos,
        np.array(
            lightpos + a2np([0, 0, -1]),
            dtype=np.float32,
        ),
        a2np([0, -1.0, 0]),
        dtype=np.float32,
    )

    return [
        px.dot(shadow_proj),
        nx.dot(shadow_proj),
        py.dot(shadow_proj),
        ny.dot(shadow_proj),
        pz.dot(shadow_proj),
        nz.dot(shadow_proj),
    ]



"""Ragdoll Object

Please note that, this is not a complete feature. Most likely many things
are still quite hard to accomplish using RagDoll. But it will be gradually
better in time. It just needs some time. Therefore, I am going to visit
this file every now and then.
"""
import math
from typing import Any, Dict, Union

from payton.scene.geometry.base import Line, Object
from payton.scene.geometry.cube import Cube
from payton.scene.geometry.sphere import Sphere
from payton.scene.material import WIREFRAME

R_COLLAR = "right_collar"
L_COLLAR = "left_collar"
R_SHOULDER = "right_shoulder"
L_SHOULDER = "left_shoulder"
NECK = "neck"
HEAD = "head"
R_HIP = "right_hip"
L_HIP = "left_hip"
R_ELBOW = "right_elbow"
L_ELBOW = "left_elbow"
R_KNEE = "right_knee"
L_KNEE = "left_knee"
R_WRIST = "right_wrist"
L_WRIST = "left_wrist"
R_FOOT = "right_foot"
L_FOOT = "left_foot"


class Joint(Sphere):
    def __init__(self, name: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.radius = 0.01
        self.material.display = WIREFRAME
        self.build_sphere()
        self.name = name


class Bone(Line):
    def __init__(
        self,
        root_joint_name: str,
        end_joint_name: str,
        length: float,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        root_joint = Joint(name=root_joint_name)
        end_joint = Joint(name=end_joint_name)
        self.add_child(root_joint_name, root_joint)
        self.root = self.children[root_joint_name]
        self.root.add_child(end_joint_name, end_joint)
        self.end = self.root.children[end_joint_name]
        self.end.position = [0.0, 0.0, length]

    def connect_to_root(self, joint: Object) -> None:
        self.root.add_child(joint.name, joint)

    def connect_to_end(self, joint: Object) -> None:
        self.end.add_child(joint.name, joint)


class RagDoll(Bone):
    def __init__(self, **kwargs: Any) -> None:
        kwargs["root_joint_name"] = "lower_spine"
        kwargs["end_joint_name"] = "higher_spine"
        kwargs["length"] = 1.0
        super().__init__(**kwargs)
        self.joints: Dict[str, Union[Joint, Object]] = {}
        self.position = [0.0, 0.0, 2.0]
        self.setup_joints()
        self.setup_geometry()

    @property
    def physics(self) -> bool:
        """TODO: Implement humanoid physics."""
        return False

    def setup_geometry(self) -> None:
        self.joints[HEAD].add_child(
            "mesh",
            Cube(from_corner=[-0.2, -0.2, 0], to_corner=[0.2, 0.2, 0.4]),
        )
        self.joints[NECK].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.2]),
        )
        self.joints[R_COLLAR].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[L_COLLAR].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[R_SHOULDER].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[L_SHOULDER].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[R_ELBOW].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.08, 0.08, 0.5]),
        )
        self.joints[L_ELBOW].add_child(
            "mesh",
            Cube(from_corner=[-0.08, -0.08, 0], to_corner=[0.08, 0.08, 0.5]),
        )
        self.joints[R_WRIST].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.01, 0], to_corner=[0.1, 0.01, 0.2]),
        )
        self.joints[L_WRIST].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.01, 0], to_corner=[0.1, 0.01, 0.2]),
        )
        self.root.add_child(
            "mesh",
            Cube(from_corner=[-0.3, -0.2, 0], to_corner=[0.3, 0.2, 1.0]),
        )
        self.joints[R_HIP].add_child(
            "mesh",
            Cube(from_corner=[-0.15, -0.15, 0], to_corner=[0.15, 0.15, 1.0]),
        )
        self.joints[L_HIP].add_child(
            "mesh",
            Cube(from_corner=[-0.15, -0.15, 0], to_corner=[0.15, 0.15, 1.0]),
        )
        self.joints[R_KNEE].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 1.0]),
        )
        self.joints[L_KNEE].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 1.0]),
        )
        self.joints[R_FOOT].add_child(
            "mesh",
            Cube(from_corner=[-0.05, -0.05, 0], to_corner=[0.05, 0.05, 0.3]),
        )
        self.joints[L_FOOT].add_child(
            "mesh",
            Cube(from_corner=[-0.08, -0.08, 0], to_corner=[0.08, 0.08, 0.3]),
        )

    def setup_joints(self) -> None:
        neck = Bone(root_joint_name="lower_neck", end_joint_name="head", length=0.2)
        self.connect_to_end(neck.root)
        head = Bone(root_joint_name="head", end_joint_name="head_top", length=0.4)
        neck.connect_to_end(head.root)
        left_collar = Bone(
            root_joint_name="l_collar_joint",
            end_joint_name="l_shoulder",
            length=0.5,
        )
        self.connect_to_end(left_collar.root)

        right_collar = Bone(
            root_joint_name="r_collar_joint",
            end_joint_name="r_shoulder",
            length=0.5,
        )
        self.connect_to_end(right_collar.root)

        right_collar.root.rotate_around_y(math.radians(100))
        left_collar.root.rotate_around_y(math.radians(-100))

        right_upper_arm = Bone(root_joint_name="r_shoulder", end_joint_name="r_elbow", length=0.5)
        right_collar.connect_to_end(right_upper_arm.root)
        right_upper_arm.rotate_around_y(math.radians(20))

        left_upper_arm = Bone(root_joint_name="l_shoulder", end_joint_name="l_elbow", length=0.5)
        left_collar.connect_to_end(left_upper_arm.root)
        left_upper_arm.rotate_around_y(math.radians(-20))

        right_lower_arm = Bone(root_joint_name="r_elbow", end_joint_name="r_wrist", length=0.5)
        right_upper_arm.connect_to_end(right_lower_arm.root)
        left_lower_arm = Bone(root_joint_name="l_elbow", end_joint_name="l_wrist", length=0.5)
        left_upper_arm.connect_to_end(left_lower_arm.root)

        right_hand = Bone(root_joint_name="r_wrist", end_joint_name="r_hand", length=0.2)
        right_lower_arm.connect_to_end(right_hand.root)
        left_hand = Bone(root_joint_name="l_wrist", end_joint_name="l_hand", length=0.2)
        left_lower_arm.connect_to_end(left_hand.root)

        right_upper_leg = Bone(root_joint_name="r_hip", end_joint_name="r_knee", length=1.0)
        self.connect_to_root(right_upper_leg.root)
        right_upper_leg.root.rotate_around_y(math.radians(160))

        left_upper_leg = Bone(root_joint_name="l_hip", end_joint_name="l_knee", length=1.0)
        self.connect_to_root(left_upper_leg.root)
        left_upper_leg.root.rotate_around_y(math.radians(-160))

        right_lower_leg = Bone(root_joint_name="r_knee", end_joint_name="r_foot", length=1.0)
        right_upper_leg.connect_to_end(right_lower_leg.root)
        left_lower_leg = Bone(root_joint_name="l_knee", end_joint_name="l_foot", length=1.0)
        left_upper_leg.connect_to_end(left_lower_leg.root)
        right_foot = Bone(root_joint_name="r_foot", end_joint_name="r_toe", length=0.3)
        right_lower_leg.connect_to_end(right_foot.root)
        right_foot.root.rotate_around_x(math.radians(90))

        left_foot = Bone(root_joint_name="l_foot", end_joint_name="l_toe", length=0.3)
        left_lower_leg.connect_to_end(left_foot.root)
        left_foot.root.rotate_around_x(math.radians(90))

        self.joints[NECK] = neck.root
        self.joints[HEAD] = head.root
        self.joints[R_COLLAR] = right_collar.root
        self.joints[L_COLLAR] = left_collar.root
        self.joints[R_SHOULDER] = right_upper_arm.root
        self.joints[L_SHOULDER] = left_upper_arm.root
        self.joints[R_ELBOW] = right_lower_arm.root
        self.joints[L_ELBOW] = left_lower_arm.root
        self.joints[R_WRIST] = right_hand.root
        self.joints[L_WRIST] = left_hand.root
        self.joints[R_HIP] = right_upper_leg.root
        self.joints[L_HIP] = left_upper_leg.root
        self.joints[R_KNEE] = right_lower_leg.root
        self.joints[L_KNEE] = left_lower_leg.root
        self.joints[R_FOOT] = right_foot.root
        self.joints[L_FOOT] = left_foot.root



# pylama:ignore=W
"""Payton Mesh Tools"""
from payton.tools.mesh.geometry import merge_mesh, subdivide
from payton.tools.mesh.line import extrude_line, lines_to_mesh, rotate_line

__all__ = [
    "extrude_line",
    "rotate_line",
    "lines_to_mesh",
    "merge_mesh",
    "subdivide",
]



from typing import Any, Dict, List, Optional

from payton.math.vector import Vector3D
from payton.scene.geometry import Object
from payton.scene.material import BASE_PARTICLE, DEFAULT, POINTS
from payton.scene.shader import PARTICLE_SHADER


class ParticleSystem(Object):
    def __init__(self, particle_scale: float = 1.0, **kwargs: Any) -> None:
        """Define a particle system

        Keyword arguments:
        particle_size -- Size of each particle
        """
        super().__init__(**kwargs)
        # Buffer to keep the state of all particles,
        # hitting update will switch buffer into particles
        self.material.texture = BASE_PARTICLE
        self.particle_scale = particle_scale
        self._previous_camera_position: Optional[Vector3D] = None
        self.meta: List[Dict[str, Any]] = []
        self.shader = PARTICLE_SHADER

    def toggle_wireframe(self) -> None:
        pass

    def track(self) -> bool:
        return False

    @property
    def physics(self) -> bool:
        """TODO: Enable particle physics for this."""
        return False

    def add(self, vertex: Vector3D, **meta: Dict[str, Any]) -> None:
        """Add a vertex/new particle to the particle system

        Note: each particle can hold a meta information as a Dict[str, Any] format.
        This is suitable to hold the energy or mass of each particle if needed.

        Keyword arguments:
        vertex -- Vertex to add to the system
        """
        self._vertices.append(vertex)
        i = len(self._vertices) - 1
        self._indices.append([i])
        self.materials[DEFAULT]._indices.append([i])
        self.materials[DEFAULT].display = POINTS
        self.meta.append(meta)
        self.refresh()



# from distutils.core import setup, find_packages
import sys

from setuptools import find_packages, setup

# For old pip versions
if sys.version_info < (3, 7):
    sys.exit("Sorry, Python < 3.7 is not supported")

setup(
    name="Payton",
    version="v0.1.8",
    author="Sinan ISLEKDEMIR",
    author_email="sinan@islekdemir.com",
    # Packages
    packages=find_packages(),
    # Include additional files into the package
    include_package_data=True,
    # Details
    url="https://github.com/sinanislekdemir/payton",
    #
    license="MIT License (MIT)",
    description="Payton 3D Kickstart Toolkit",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=">3.7",
    # Dependent packages (distributions)
    install_requires=[
        "numpy>=1.18.2",
        "Pillow>=8.2.0",
        "PyOpenGL>=3.1.3rc1",
        "pyrr>=0.10.3",
        "PySDL2>=0.9.7",
    ],
)



"""Basic Window and UI Elements Support

This is a simple support but works without much hassle.
I tried to make the whole mechanism as basic as possible.
Anyone without any UI coding experience should be able to get started
with the basic stuff"""

from enum import Enum
from typing import Any, Callable, Dict, List, Optional

import numpy as np

from payton.math.vector import Vector3D
from payton.scene.geometry.base import Object
from payton.scene.gui.base import Shape2D, Text
from payton.scene.shader import Shader


class Theme:
    """User interface theme"""

    def __init__(self) -> None:
        """Initialize the theme"""
        self.opacity = 0.9

        self.background_color: Vector3D = [0.05, 0.05, 0.05]
        self.text_color: Vector3D = [1.0, 1.0, 1.0]
        self.title_background_color: Vector3D = [
            235 / 255,
            210 / 255,
            52 / 255,
        ]
        self.title_text_color: Vector3D = [0.0, 0.0, 0.0]

    def secondary(self) -> None:
        """Secondary colors for the theme"""
        self.background_color, self.text_color = self.title_background_color, self.title_text_color
        self.title_background_color, self.title_text_color = self.background_color, self.text_color


class WindowAlignment(Enum):
    """Window alignment on the HUD"""

    FREE = "free"
    LEFT = "left"
    TOP = "top"
    RIGHT = "right"
    BOTTOM = "bottom"


class WindowElement(Shape2D):
    """Window Base Element"""

    def __init__(
        self,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Window Element

        Keyword arguments:
        width -- Width of the window element
        height -- Height of the window element
        left -- Left position of the window element
        top -- Top position of the window element
        align -- Alignment of the element
        theme -- Theme to be used
        """
        kwargs["position"] = (left, top, 0)
        kwargs["size"] = (width, height)
        super().__init__(**kwargs)
        self._init: bool = False
        self.align = align
        self.theme = theme if theme is not None else Theme()
        self.material.opacity = self.theme.opacity

    def _reposition(self) -> None:
        if self.align == WindowAlignment.LEFT:
            self.position = [0.0, 0.0, 0.0]
            self.size = [self.size[0], self._parent_height]
        elif self.align == WindowAlignment.RIGHT:
            self.position = [self._parent_width - self.size[0], 0.0, 0.0]
            self.size = [self.size[0], self._parent_height]
        elif self.align == WindowAlignment.TOP:
            self.position = [0.0, 0.0, 0.0]
            self.size = [self._parent_width, self.size[1]]
        elif self.align == WindowAlignment.BOTTOM:
            self.position = [0, self._parent_height - self.size[1], 0]
            self.size = [self._parent_width, self.size[1]]

    def draw(self) -> None:
        self._reposition()

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ) -> None:
        if not self._init:
            self.draw()

        super().render(lit, shader, parent_matrix)

    def add_child(self, name: str, obj: Object) -> bool:
        """Add a window element as a child

        Keyword arguments:
        name -- Name of the window element to be added
        obj -- Object to add
        """
        if not isinstance(obj, Shape2D):
            return False
        if obj.position[0] > self.size[0]:
            obj.position[0] = self.size[0] - 1
        total_v = obj.position[0] + obj.size[0]
        exceed = int(total_v - self.size[0])
        if exceed > 0:
            obj.size = [obj.size[0] - exceed, obj.size[1]]

        if obj.position[1] > self.size[1]:
            obj.position[1] = self.size[1] - 1
        total_v = obj.position[1] + obj.size[1]
        exceed = int(total_v - self.size[1])
        if exceed > 0:
            obj.size = [obj.size[0], obj.size[1] - exceed]

        super().add_child(name, obj)
        obj._set_parent_size(self.size[0], self.size[1])
        return True


class Window(WindowElement):
    """Main window frame"""

    def __init__(
        self,
        title: str = "",
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs: Dict[str, Any],
    ):
        """Initialize Window Frame

        @NOTE: I am aware that I could have just define "title" as other
               elemenets are already arguments of WindowElement but some
               code hinting and completion tools are a bit stupid. They
               can not inherit from parent class. So this is a ide-candy

        Keyword arguments:
        title -- Title of the window
        width -- Width of the window
        height -- Height of the window
        left -- Left position of the window
        top -- Top position of the window
        align -- Alignment of the
        theme -- Theme to be used
        """
        super().__init__(
            width=width,
            height=height,
            left=left,
            top=top,
            align=align,
            theme=theme,
            **kwargs,
        )
        self.title = title
        self.add_child(
            "title",
            Text(
                position=[0, 0, 0],
                size=[self.size[0], 20],
                label=self.title,
                # bgcolor=self.theme.title_background_color,
                color=self.theme.title_text_color,
                opacity=0.5,
            ),
        )

    def draw(self) -> None:
        """Create the frame polygons"""
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 22, 1], [w, h, 1], [w, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[0, 22, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[0, 0, 1], [w - 1, 22, 1], [w - 1, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, 22, 1], [w - 1, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self._init = True


class Panel(WindowElement):
    """Panel element"""

    def draw(self) -> None:
        """Create the panel polygons"""
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 0, 1], [w, h, 1], [w, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[1, 1, 1], [w - 1, h - 1, 1], [w - 1, 1, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[1, 1, 1], [1, h - 1, 1], [w - 1, h - 1, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )

            self._init = True


class Button(Panel):
    def __init__(
        self,
        label: str,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        on_click: Optional[Callable] = None,
        **kwargs: Any,
    ):
        """Initialize the Button element

        label -- Label of the button
        width -- Width of the button
        height -- Height of the button
        left -- Left position of the button
        top -- Top position of the button
        align -- Alignment of the button
        theme -- Theme to be used
        on_click -- On Click callback method
        """
        kwargs["on_click"] = on_click
        super().__init__(
            width=width,
            height=height,
            left=left,
            top=top,
            align=align,
            theme=theme,
            **kwargs,
        )
        height = max(height, 30)

        self._label = label
        self.text = Text(
            position=[0, 0, 1],
            size=[10, 10],
            label=label,
            color=self.theme.text_color,
        )
        self.add_child("label", self.text)

    def draw(self, **kwargs: Any) -> None:
        """Create the button polygons"""
        super().draw()
        text_size = self.text.text_size
        x = (self.size[0] - text_size[0]) / 2
        y = (self.size[1] - text_size[1]) / 2
        x = max(x, 0)
        y = max(y, 0)
        y -= 4
        self.text.label = self._label
        self.text.position = [x, y]
        crop: List[int] = [0, 0, text_size[0], text_size[1] + 4]
        if text_size[0] > self.size[0]:
            crop[0] = int((text_size[0] - self.size[0]) / 2.0)
            crop[2] = int(text_size[0] - crop[0])
            diff = int(crop[2] - crop[0])
            text_size = (diff, text_size[1])
        if text_size[1] > self.size[1]:
            crop[1] = int((text_size[1] - self.size[1]) / 2.0)
            crop[3] = int(text_size[1] - crop[1])
            text_size = (text_size[0], self.size[1])
        self.text.crop = crop
        self.text.size = [int(text_size[0]), int(text_size[1] + 4)]

    @property
    def label(self) -> str:
        """Get the button label"""
        return self._label

    @label.setter
    def label(self, text: str) -> None:
        """Set the button label

        Keyword arguments:
        text -- Text of the label
        """
        self._label = text
        self.text.label = self._label


class EditBox(Panel):
    def __init__(
        self,
        value: str,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        on_change: Optional[Callable] = None,
        multiline: bool = False,
        **kwargs: Any,
    ):
        """Initialize the edit box element

        value -- Default value of the editbox
        width -- Width of the editbox
        height -- Height of the editbox
        left -- Left position of the editbox
        top -- Top position of the editbox
        align -- Alignment of the editbox
        theme -- Theme to be used
        on_change -- On change callback method (expects a string argument) `def on_change(text: str)`
        """
        kwargs["on_click"] = self._focus
        height = max(height, 30)

        super().__init__(
            width=width,
            height=height,
            left=left,
            top=top,
            align=align,
            theme=theme,
            **kwargs,
        )

        self.multiline = multiline
        self.theme.secondary()
        self.on_change = on_change
        self._value = value
        self.text = Text(
            position=[0, 0, 0],
            size=[width, height],
            label=self._value,
            color=self.theme.text_color,
        )
        self.add_child("label", self.text)
        self._cursor = -1

    def _on_keypress(self, instr: str) -> None:
        new = self.value[0 : self._cursor] + instr + self.value[self._cursor :]
        self._cursor = len(self.value[0 : self._cursor] + instr)
        self.value = new

    def _focus(self) -> None:
        self._cursor = len(self.value)
        self._init = False

    def cursor_left(self) -> None:
        """Move the cursor to the left"""
        self._cursor -= 1
        self._cursor = max(self._cursor, 0)
        self._init = False

    def cursor_right(self) -> None:
        """Move the cursor to the right"""
        self._cursor += 1
        self._cursor = min(self._cursor, len(self.value))
        self._init = False

    def backspace(self) -> None:
        """Delete the previous character"""
        if self._cursor == 0:
            return
        self.value = self.value[0 : self._cursor - 1] + self.value[self._cursor :]
        self.cursor_left()

    def _exit(self) -> None:
        self._cursor = -1
        if callable(self.on_change):
            self.on_change(self._value)
        self._init = False

    def draw(self, **kwargs: Any) -> None:
        """Create the editbox polygons and texture along with cropping"""
        super().draw()
        if self._cursor > -1:
            label = self._value[0 : self._cursor] + "|" + self._value[self._cursor :]
        else:
            label = self._value
        self.text.label = label
        if self.multiline:
            self.text.wrap(self.size[0])
        text_size = list(self.text.text_size)
        x = 1
        y = (self.size[1] - text_size[1]) / 2
        y = max(y, 0)
        y += 4
        self.text.position = [x, y]
        crop = [0, 0, text_size[0], text_size[1] + 4]
        if text_size[0] > self.size[0]:
            crop[0] = text_size[0] - self.size[0]
            crop[2] = text_size[0]

        if text_size[1] > self.size[1]:
            crop[1] = text_size[1] - self.size[1]
            crop[3] = self.text.text_size[1]

        self.text.crop = crop
        self.text._init = False
        self.text._init_text = False

    @property
    def value(self) -> str:
        """Return the editbox value as a string"""
        return self._value

    @value.setter
    def value(self, text: str) -> None:
        """Set the editbox value

        Keyword arguments:
        text -- Text value to set
        """
        self._value = text
        self.text.label = self._value
        if self.text.text_size[0] > self.size[0] and self.multiline:
            self.text.wrap(self.size[0])
        self.refresh()
        self._init = False
